/*
libfawk - A function-only AWK dialect

Copyright (c) 2017..2020 Tibor 'Igor2' Palinkas. All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

  * 1. Redistributions of source code must retain the above copyright notice,
    this list of conditions and the following disclaimer.
  * 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the distribution.
  * 3. Neither the name of the copyright holder nor the names of its
    contributors may be used to endorse or promote products derived from this
    software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

Project page: http://repo.hu/projects/libfawk
Source code: svn://repo.hu/libfawk/trunk
Contact the author: http://igor2.repo.hu/contact.html
*/

#include <stdlib.h>
#include <string.h>
#include "libfawk.h"
#include "execute.h"
#include "cast.h"
#include "str.h"
#include "parser.h"
#include "array.h"

#define STACKA(addr) ((ctx->stack.page[(addr) / FAWK_STACK_PAGE_SIZE])[((addr) % FAWK_STACK_PAGE_SIZE)])
#define STACKR(offs) STACKA(ctx->sp + (offs))

#define NEED_STACK(entries) fawk_assert((ctx->sp-(entries)) >= ctx->fp);

#define FAWK_CAST_TO_NUM(cell) \
do { \
	if (cell->type != FAWK_NUM) \
		fawk_cast_to_num(ctx, cell); \
} while(0)

#define FAWK_CAST_TO_STR(cell) \
do { \
	if (cell->type != FAWK_STR) \
		fawk_cast_to_str(ctx, cell); \
} while(0)

/* should be inlined for performance */
static void cell_free(fawk_ctx_t *ctx, fawk_cell_t *cell)
{
	int n;
	switch(cell->type) {
		case FAWK_STR:
		case FAWK_STRNUM:
			if (cell->data.str != NULL)
				fawk_str_free(ctx, cell->data.str);
			break;
		case FAWK_ARRAY:
			fawk_array_free(ctx, cell);
			return;
		case FAWK_SYMREF:
			for(n = 0; (n < cell->data.symref.idx_len) && (cell->data.symref.idx_len != -1); n++)
				if ((cell->data.symref.idx[n].type == FAWK_STR) || (cell->data.symref.idx[n].type == FAWK_STRNUM))
					fawk_str_free(ctx, cell->data.symref.idx[n].data.str);
			fawk_free(ctx, cell->data.symref.idx);
			break;
		default: break;
	}
	cell->type = FAWK_NIL;
}

FAWK_API void fawk_cell_free(fawk_ctx_t *ctx, fawk_cell_t *cell)
{
	cell_free(ctx, cell);
}

static void cellcpy(fawk_ctx_t *ctx, fawk_cell_t *dst, const fawk_cell_t *src)
{
	cell_free(ctx, dst);
	*dst = *src;
	switch(src->type) {
		case FAWK_STR:
		case FAWK_STRNUM:
			dst->data.str = fawk_str_dup(ctx, src->data.str); if (dst->data.str == NULL) dst->type = FAWK_NIL;
			break;
		case FAWK_ARRAY:
			dst->type = FAWK_ARRAY;
			dst->data.arr = src->data.arr;
			dst->data.arr->refco++;
		default: break;
	}
}

FAWK_API void fawk_cell_cpy(fawk_ctx_t *ctx, fawk_cell_t *dst, const fawk_cell_t *src)
{
	cellcpy(ctx, dst, src);
}

/* should be inlined for performance */
static fawk_cell_t *push_alloc(fawk_ctx_t *ctx)
{
	fawk_cell_t *res;
	if (ctx->stack.avail == 0) { /* have to open a new page */
		if (ctx->stack.used >= ctx->stack.alloced) { /* need to grow page index */
			fawk_cell_t **pg;
			ctx->stack.alloced += 128;
			if ((pg = fawk_realloc(ctx, ctx->stack.page, sizeof(fawk_cell_t *) * ctx->stack.alloced)) == NULL) { ctx->stack.alloced = 0; ctx->exec.error = 1; return NULL; }
			ctx->stack.page = pg;
		}
		ctx->stack.page[ctx->stack.used] = fawk_malloc(ctx, sizeof(fawk_cell_t) * FAWK_STACK_PAGE_SIZE);
		if (ctx->stack.page[ctx->stack.used] == NULL) { ctx->exec.error = 1; return NULL;}
		ctx->stack.avail = FAWK_STACK_PAGE_SIZE; ctx->stack.used++;
	}
	ctx->stack.avail--;
	res = &STACKA(ctx->sp);
	res->name = NULL;
	res->type = FAWK_NIL;
	ctx->sp++;
	return res;
}

FAWK_API void fawk_reset(fawk_ctx_t *ctx)
{
	size_t n;
	for(n = 0; n < ctx->sp; n++)
		cell_free(ctx, &STACKA(n));
	ctx->ip = ctx->fp = ctx->sp = 0;
	fawk_free(ctx, ctx->errbuff);
	ctx->errbuff = NULL;
	ctx->errbuff_alloced = 0;
}


FAWK_API fawk_cell_t *fawk_push_alloc(fawk_ctx_t *ctx)
{
	return push_alloc(ctx);
}

FAWK_API size_t fawk_push_num(fawk_ctx_t *ctx, fawk_num_t num)
{
	fawk_cell_t *cell = push_alloc(ctx);
	if (cell == NULL)
		return FAWK_STACK_INVALID;
	cell->type = FAWK_NUM;
	cell->data.num = num;
	return ctx->sp-1;
}

FAWK_API size_t fawk_push_str(fawk_ctx_t *ctx, const char *str)
{
	fawk_cell_t *cell = push_alloc(ctx);
	if (cell == NULL)
		return FAWK_STACK_INVALID;
	cell->data.str = fawk_str_new_from_literal(ctx, str, -1);
	cell->type = (cell->data.str == NULL) ? FAWK_NIL : FAWK_STR;
	return ctx->sp-1;
}


#define PUSH() push_alloc(ctx)

#define POP() \
do { \
	cell_free(ctx, &STACKR(-1)); \
	ctx->sp--; \
	ctx->stack.avail++; \
} while(0)

FAWK_API fawk_num_t fawk_pop_num(fawk_ctx_t *ctx, int expect_num)
{
	fawk_cell_t *cell;

	NEED_STACK(1);
	cell = &STACKR(-1);
	if (cell->type != FAWK_NUM) {
		if (expect_num) {
			fawk_assert(cell->type == FAWK_NUM);
			POP();
			return 0;
		}
		fawk_cast_to_num(ctx, cell);
	}

	POP();
	return cell->data.num;
}

FAWK_API int fawk_pop(fawk_ctx_t *ctx, fawk_cell_t *dst)
{
	if (ctx->sp < 0)
		return -1;
	*dst = STACKR(-1);
	STACKR(-1).type = FAWK_NIL;
	POP();
	return 0;
}

FAWK_API fawk_cell_t *fawk_peek(fawk_ctx_t *ctx, int addr)
{
	return (addr >= 0) ? &STACKA(addr) : &STACKR(addr);
}

/* Defined here for inlining */
static fawk_cell_t *symtab_deref(fawk_ctx_t *ctx, const fawk_symref_t *sr, int arr_create, fawk_cell_t **parent)
{
	fawk_cell_t *base, *child;
	int n;

	/* base is the initial cell: either a local var or a global var, either scalar or array */
	base = (sr->is_local) ? &STACKA(ctx->fp + sr->ref.local) : sr->ref.global;
	fawk_assert(base != NULL);

	if (parent != NULL)
		*parent = NULL;

	if (sr->idx_len == 0) /* scalar -> we are done */
		return base;

	for(n = 0; (n < sr->idx_len) && (sr->idx_len != -1); n++) { 	/* do the array lookup */
		if (base->type == FAWK_NIL)
			fawk_array_init(ctx, base);
		else if (base->type != FAWK_ARRAY) {
			FAWK_ERROR(ctx, 64, (FAWK_ERR, "deref: symbol is not an array but is indexed like if it was\n"));
			return NULL;
		}
		if ((child = htpp_get(&base->data.arr->hash, &sr->idx[n])) == NULL) {
			fawk_arridx_t *idx;
			if (!arr_create)
				return NULL;
			child = fawk_malloc(ctx, sizeof(fawk_cell_t)); if (child == NULL) { return NULL; }
			child->type = FAWK_NIL;
			idx = fawk_malloc(ctx, sizeof(fawk_arridx_t)); if (idx == NULL) { free(child); return NULL; }
			idx->type = sr->idx[n].type;
			if ((sr->idx[n].type == FAWK_STR) || (sr->idx[n].type == FAWK_STRNUM)) {
				idx->data.str = fawk_str_dup(ctx, sr->idx[n].data.str);
				if (idx->data.str == NULL) { idx->type = FAWK_NIL; FAWK_ERROR(ctx, 64, (FAWK_ERR, "memory exhausted\n")); }
			}
			else
				idx->data.num = sr->idx[n].data.num;
			htpp_set(&base->data.arr->hash, idx, child);
		}
		if ((n < sr->idx_len-1) && (child->type == FAWK_NIL))
			fawk_array_init(ctx, child);
		if (parent != NULL)
			*parent = base;
		base = child;
	}
	return base;
}

FAWK_API fawk_cell_t *fawk_symtab_deref(fawk_ctx_t *ctx, const fawk_symref_t *sr, int arr_create, fawk_cell_t **parent)
{
	return symtab_deref(ctx, sr, arr_create, parent);
}


/* deref the symref on the top of the stack and push the value;
   if and_pop is non-zero, remove the symref; returns the dereferenced source cell */
static fawk_cell_t *topvar(fawk_ctx_t *ctx, int and_pop)
{
	fawk_cell_t *cdst, *csrc, *cell;
	NEED_STACK(1);
	cell = &STACKR(-1);
	fawk_assert(cell->type == FAWK_SYMREF);
	if ((csrc = symtab_deref(ctx, &cell->data.symref, 1, NULL)) == NULL)
		return NULL;
	if (and_pop)
		POP();
	cdst = PUSH(); if (cdst == NULL) return NULL;
	cellcpy(ctx, cdst, csrc);
	return csrc;
}

static void exec_call(fawk_ctx_t *ctx, int argc)
{
	fawk_cell_t *fc, *nil, *vararg, *child, vtmp;

	fc = &STACKR(-(argc + 1));
	fawk_assert(fc->type == FAWK_SYMREF);
	if ((fc = symtab_deref(ctx, &fc->data.symref, 1, NULL)) == NULL)
		return;

	if (fc->type != FAWK_FUNC) {
		FAWK_ERROR(ctx, 64, (FAWK_ERR, "can't call: symbol is not a function\n"));
		return;
	}

	if (fc->data.func.cfunc == NULL) { /* good old fawk function */
		/* if there is a vararg part, convert the non-fixed part to an array and sotre it as local VARARG[] */
		if (fc->data.func.numfixedargs >= 0) {
			int vac = argc - fc->data.func.numfixedargs - 1;
			fawk_array_init(ctx, &vtmp);
			while(argc > fc->data.func.numfixedargs) {
				fawk_arridx_t *idx = malloc(sizeof(fawk_arridx_t)); if (idx == NULL) goto enomem;
				idx->type = FAWK_NUM; idx->data.num = vac--;
				child = fawk_malloc(ctx, sizeof(fawk_cell_t)); if (child == NULL) { enomem:; fawk_cell_free(ctx, &vtmp); return; }
				*child = STACKR(-1); ctx->sp--; ctx->stack.avail++; argc--;
				htpp_set(&vtmp.data.arr->hash, idx, child);
			}
			vararg = PUSH(); *vararg = vtmp;
		}
		else if (argc > fc->data.func.numargs) {FAWK_ERROR(ctx, 64 + strlen(fc->data.func.name), (FAWK_ERR, "Function '%s' called with more arguments than it takes\n", fc->data.func.name)); return; }

		/* fill up missing arguments with NILs */
		while(argc < fc->data.func.numargs) {
			nil = PUSH(); if (nil == NULL) {FAWK_ERROR(ctx, 64, (FAWK_ERR, "memory exhausted\n")); return; }
			nil->type = FAWK_NIL;
			argc++;
		}

		fawk_push_num(ctx, ctx->fp);
		fawk_push_num(ctx, ctx->ip+1);
		ctx->fp = ctx->sp;
		ctx->ip = fc->data.func.ip-1;
	}
	else { /* a function implemented in C */
		fawk_cell_free(ctx, &STACKR(-(argc + 1)));
		fc->data.func.cfunc(ctx, fc->data.func.name, argc, &STACKR(-(argc + 1)));
		for(;argc > 0; argc--)
			POP();
		ctx->ip++;
	}
}

static int idx_steal_cell(fawk_arridx_t *res, fawk_cell_t *csrc, int allow_nil)
{
	res->type = csrc->type;
	switch(csrc->type) {
		case FAWK_NUM: res->data.num = csrc->data.num; return 0;
		case FAWK_STRNUM:
		case FAWK_STR:
			res->data.str = csrc->data.str; /* steal the string from the stack slot */
			csrc->data.str = NULL;
			return 0;
		case FAWK_NIL: if (allow_nil) return 0;
		default: break;
	}
	return -1;
}

/*** Execute helper macros ***/

/* instruction switch "break;" and bump IP by the number of params */
#define BREAK(num_param) ctx->ip += num_param; break

/* Execute action on the top of the stack */
#define UNOP_ON_TOP(action) \
	NEED_STACK(1); \
	cell = &STACKR(-1); \
	FAWK_CAST_TO_NUM(cell); \
	action

/* load cell pointers op1 and op2 to point to the top and top-1st operands
   DOES NOT POP */
#define BINOP(op1, op2) \
do { \
	NEED_STACK(2); \
	op1 = &STACKR(-1); \
	op2 = &STACKR(-2); \
} while(0)

/* same as BINOP() but also cast both ops to num; pops one operand */
#define BINOP_NUM(op1, op2) \
do { \
	BINOP(op1, op2); \
	FAWK_CAST_TO_NUM(op1); \
	FAWK_CAST_TO_NUM(op2); \
	POP(); \
} while(0)

/* same as BINOP() but also cast both ops to num; pops one operand; returns -2 if any op is nil */
#define BINOP_STRNUM(sres, op1, op2) \
do { \
	BINOP(op1, op2); \
	if ((i->data.ins == FAWKI_EQ) && (op1->type == FAWK_NIL) && (op2->type == FAWK_NIL)) { STACKR(-2).type = FAWK_NUM; sres = 0; } \
	else if ((i->data.ins == FAWKI_NEQ) && (((op1->type == FAWK_NIL) && (op2->type != FAWK_NIL)) || ((op2->type == FAWK_NIL) && (op1->type != FAWK_NIL))))  { STACKR(-2).type = FAWK_NUM; sres = 1; } \
	else if ((op1->type == FAWK_NIL) || (op2->type == FAWK_NIL)) { STACKR(-2).type = FAWK_NUM; sres = -2; }\
	else if ((op1->type == FAWK_STR) || (op2->type == FAWK_STR)) { \
		FAWK_CAST_TO_STR(op1); \
		FAWK_CAST_TO_STR(op2); \
		sres = strcmp(op2->data.str->str, op1->data.str->str); \
		STACKR(-2).type = FAWK_NUM; \
	} \
	else { \
		FAWK_CAST_TO_NUM(op1); \
		FAWK_CAST_TO_NUM(op2); \
		if (op2->data.num > op1->data.num) sres = 1; \
		else if (op2->data.num == op1->data.num) sres = 0; \
		else sres = -1; \
	} \
	POP(); \
} while(0)


FAWK_API fawk_execret_t fawk_execute(fawk_ctx_t *ctx, size_t steps)
{
	fawk_cell_t *cell, *csrc, *cdst;
	int n;

	for(;steps > 0; steps--) {
		fawk_code_t *i = &ctx->code.code[ctx->ip];

		if (ctx->exec.error)
			return FAWK_ER_ERROR;

		if (ctx->exec.trace)
			printf("[%ld] %x\n", (long)ctx->ip, i->data.ins);

		fawk_assert(i[0].type == FAWKC_INS);

		switch(i->data.ins) {
			case FAWKI_MAKE_SYMREF:
				{
					fawk_cell_t res;
					fawk_assert(i[1].type == FAWKC_SYMREF);
					fawk_assert(i[2].type == FAWKC_NUM);
					res.type = FAWK_SYMREF;
					res.data.symref.is_local = i[1].data.symref->is_local;
					res.data.symref.idx_len = i[2].data.num;

					if (i[1].data.symref->is_local) {
						res.name = "<local>";
						res.data.symref.ref.local = i[1].data.symref->ref.local;
						if ((res.data.symref.idx_len > 0) || (res.data.symref.idx_len == -1)) {
							fawk_cell_t *base = &STACKA(ctx->fp + res.data.symref.ref.local);
							if (base->type == FAWK_NIL) fawk_array_init(ctx, base);
						}
					}
					else {
						res.name = i[1].data.symref->ref.global->name;
						res.data.symref.ref.global = i[1].data.symref->ref.global;
					}

					/* allocate array index list and pop them from the stack; -1 may be special if size_t is unsigned */
					if ((res.data.symref.idx_len > 0) && (res.data.symref.idx_len != -1)) {
						int n, d;
						res.data.symref.idx = fawk_malloc(ctx, sizeof(fawk_arridx_t) * res.data.symref.idx_len); if (res.data.symref.idx == NULL) return FAWK_ER_ERROR;
						for(n = 0, d = res.data.symref.idx_len-1; n < res.data.symref.idx_len; n++,d--) {
							if (idx_steal_cell(&(res.data.symref.idx[d]), &STACKR(-1), 1) != 0)
								abort(); /* can't get here without an internal compiler error */
							POP();
						}
					}
					else
						res.data.symref.idx = NULL;
					cell = PUSH(); if (cell == NULL) return FAWK_ER_ERROR;
					*cell = res;
				}
				BREAK(2);
			case FAWKI_PUSH_NUM:
				fawk_assert(i[1].type == FAWKC_NUM);
				cell = PUSH(); if (cell == NULL) return FAWK_ER_ERROR;
				cell->type = FAWK_NUM;
				cell->data.num = i[1].data.num;
				BREAK(1);
			case FAWKI_PUSH_STR:
				fawk_assert(i[1].type == FAWKC_STR);
				cell = PUSH(); if (cell == NULL) return FAWK_ER_ERROR;
				cell->data.str = fawk_str_dup(ctx, i[1].data.str);
				cell->type = (cell->data.str == NULL) ? FAWK_NIL : FAWK_STR;
				BREAK(1);
			case FAWKI_PUSH_SYMVAL:
				topvar(ctx, 1);
				BREAK(0);
			case FAWKI_PUSH_TOPVAR:
				topvar(ctx, 0);
				BREAK(0);
			case FAWKI_PUSH_REL:
				cell = PUSH(); if (cell == NULL) return FAWK_ER_ERROR;
				fawk_cell_cpy(ctx, cell, &STACKR(((int)i[1].data.num)-1));
				BREAK(1);
			case FAWKI_PUSH_NIL:
				cell = PUSH(); if (cell == NULL) return FAWK_ER_ERROR;
				cell->type = FAWK_NIL;
				BREAK(0);
			case FAWKI_SET:
				csrc = &STACKR(-1);
				cell = &STACKR(-2);
				fawk_assert(cell->type == FAWK_SYMREF);
				if ((cdst = symtab_deref(ctx, &cell->data.symref, 1, NULL)) == NULL)
					return FAWK_ER_ERROR;
				cellcpy(ctx, cdst, csrc);
				POP();
				BREAK(0);
			case FAWKI_POP:
				NEED_STACK(1);
				POP();
				BREAK(0);
			case FAWKI_POPJNZ:
				NEED_STACK(1);
				cell = &STACKR(-1);
				FAWK_CAST_TO_NUM(cell);
				if (cell->data.num != 0.0) {
					POP();
					goto exec_jump;
				}
				POP();
				BREAK(1);
			case FAWKI_POPJZ:
				NEED_STACK(1);
				cell = &STACKR(-1);
				FAWK_CAST_TO_NUM(cell);
				if (cell->data.num == 0.0) {
					POP();
					goto exec_jump;
				}
				POP();
				BREAK(1);
			case FAWKI_NEG:
				UNOP_ON_TOP(cell->data.num = -cell->data.num);
				BREAK(0);
			case FAWKI_NOT:
				UNOP_ON_TOP(cell->data.num = (cell->data.num == 0.0));
				BREAK(0);
			case FAWKI_EQ:
				BINOP_STRNUM(n, csrc, cdst);
				cdst->data.num = (n != -2) && (n == 0);
				BREAK(0);
			case FAWKI_NEQ:
				BINOP_STRNUM(n, csrc, cdst);
				cdst->data.num = (n != -2) && (n != 0);
				BREAK(0);
			case FAWKI_LTEQ:
				BINOP_STRNUM(n, csrc, cdst);
				cdst->data.num = (n != -2) && (n <= 0);
				BREAK(0);
			case FAWKI_GTEQ:
				BINOP_STRNUM(n, csrc, cdst);
				cdst->data.num = (n != -2) && (n >= 0);
				BREAK(0);
			case FAWKI_IN:
				{
					fawk_arridx_t resi;

					NEED_STACK(2);
					csrc = &STACKR(-1);
					fawk_assert(csrc->type == FAWK_SYMREF);
					csrc = fawk_symtab_deref(ctx, &csrc->data.symref, 0, &cell);
					if (csrc->type != FAWK_ARRAY) {
						FAWK_ERROR(ctx, 64, (FAWK_ERR, "in: symbol is not an array, can't interpret 'in' it\n"));
						return FAWK_ER_ERROR;
					}
					POP();
					cdst = &STACKR(-1);

					if (idx_steal_cell(&resi, cdst, 0) != 0) {
						FAWK_ERROR(ctx, 64, (FAWK_ERR, "in: invalid index type\n"));
						return FAWK_ER_ERROR;
					}

					n = htpp_has(&csrc->data.arr->hash, &resi);
					cell_free(ctx, cdst);
					cdst->type = FAWK_NUM;
					cdst->data.num = n;
				}
				BREAK(0);
			case FAWKI_LT:
				BINOP_STRNUM(n, csrc, cdst);
				cdst->data.num = (n != -2) && (n < 0);
				BREAK(0);
			case FAWKI_GT:
				BINOP_STRNUM(n, csrc, cdst);
				cdst->data.num = (n != -2) && (n > 0);
				BREAK(0);
			case FAWKI_ADD:
				BINOP_NUM(csrc, cdst);
				cdst->data.num += csrc->data.num;
				BREAK(0);
			case FAWKI_SUB:
				BINOP_NUM(csrc, cdst);
				cdst->data.num -= csrc->data.num;
				BREAK(0);
			case FAWKI_MUL:
				BINOP_NUM(csrc, cdst);
				cdst->data.num *= csrc->data.num;
				BREAK(0);
			case FAWKI_DIV:
				BINOP_NUM(csrc, cdst);
				cdst->data.num /= csrc->data.num;
				BREAK(0);
			case FAWKI_MOD:
				BINOP_NUM(csrc, cdst);
				cdst->data.num = fawk_fmod(cdst->data.num, csrc->data.num);
				BREAK(0);
			case FAWKI_INCDEC:
				{
					fawk_num_t diff;
					fawk_assert(i[1].type == FAWKC_NUM);
					diff = (((int)i[1].data.num) & FAWK_INCDEC_INC) ? +1 : -1;
					csrc = topvar(ctx, 1);
					cdst = &STACKR(-1);
					FAWK_CAST_TO_NUM(csrc);
					if (!(((int)i[1].data.num) & FAWK_INCDEC_POST)) {
						FAWK_CAST_TO_NUM(cdst);
						cdst->data.num += diff;
					}
					csrc->data.num += diff;
					BREAK(1);
				}
			case FAWKI_CONCAT:
				{
					fawk_str_t *ssrc, *sdst;
					NEED_STACK(2);
					csrc = &STACKR(-1);
					cdst = &STACKR(-2);
					FAWK_CAST_TO_STR(csrc); if ((csrc->type != FAWK_STR) && (csrc->type != FAWK_STRNUM)) return FAWK_ER_ERROR;
					FAWK_CAST_TO_STR(cdst); if ((cdst->type != FAWK_STR) && (cdst->type != FAWK_STRNUM)) return FAWK_ER_ERROR;
					/* steal the strings of stack slots */
					ssrc = csrc->data.str;
					sdst = cdst->data.str;
					cdst->data.str = csrc->data.str = NULL;
					POP();
					POP();
					cell = PUSH(); if (cell == NULL) return FAWK_ER_ERROR;
					cell->data.str = fawk_str_concat(ctx, sdst, ssrc);
					cell->type = (cell->data.str == NULL) ? FAWK_NIL : FAWK_STR;
				}
				BREAK(0);
			case FAWKI_FORIN_FIRST:
				{
					fawk_arridx_t *list;
					size_t len;

					cdst = &STACKR(-1);
					fawk_assert(cdst->type == FAWK_SYMREF);
					csrc = symtab_deref(ctx, &cdst->data.symref, 0, NULL);
					list = fawk_array_dump_list(ctx, csrc, &len);
					if (list == NULL) { FAWK_ERROR(ctx, 64, (FAWK_ERR, "for-in: not an array\n")); return FAWK_ER_ERROR; }
					cell_free(ctx, cdst);
					cdst->type = FAWK_SYMREF;
					cdst->data.symref.is_local = 0;
					cdst->data.symref.ref.global = NULL;
					cdst->data.symref.idx_len = len;
					cdst->data.symref.idx = list;
				}
				BREAK(0);
			case FAWKI_FORIN_NEXT:
				{
					fawk_assert(i[1].type == FAWKC_NUM);
					csrc = &STACKR(-1);
					fawk_assert(csrc->type == FAWK_SYMREF);
					if ((csrc->data.symref.idx_len > 0) && (csrc->data.symref.idx_len != -1)) {

						/* set cdst to point to the target loop var */
						cdst = &STACKR(-2);
						fawk_assert(cdst->type == FAWK_SYMREF);
						if ((cdst = symtab_deref(ctx, &cdst->data.symref, 1, NULL)) == NULL)
							return FAWK_ER_ERROR;
						cell_free(ctx, cdst);

						csrc->data.symref.idx_len--;
						if ((csrc->data.symref.idx[csrc->data.symref.idx_len].type == FAWK_STR) || (csrc->data.symref.idx[csrc->data.symref.idx_len].type == FAWK_STRNUM)) {
/*							printf("--FORIN %s\n", csrc->data.symref.idx[csrc->data.symref.idx_len].data.str->str);*/
							cdst->type = FAWK_STR;
							cdst->data.str = csrc->data.symref.idx[csrc->data.symref.idx_len].data.str;
							csrc->data.symref.idx[csrc->data.symref.idx_len].data.str = NULL;
						}
						else if (csrc->data.symref.idx[csrc->data.symref.idx_len].type == FAWK_NIL) cdst->type = FAWK_NIL;
						else {
/*							printf("--FORIN "FAWK_NUM_PRINTF_FMT"\n", csrc->data.symref.idx[csrc->data.symref.idx_len].data.num);*/
							cdst->type = FAWK_NUM;
							cdst->data.num = csrc->data.symref.idx[csrc->data.symref.idx_len].data.num;
						}
						/* jump back for the next iteration */
						ctx->ip = i[1].data.num - 1;
						break;
					}
					else { /* exit the loop */
						POP();
						POP();
					}
				}
				BREAK(1);
			case FAWKI_CALL:
				fawk_assert(i[1].type == FAWKC_NUM);
				exec_call(ctx, (int)i[1].data.num);
				BREAK(0);
			case FAWKI_RET:
				{
					int numargs, n;
					fawk_cell_t retval = STACKR(-1); /* steal it from the stack */
					ctx->sp--; ctx->stack.avail++;

					fawk_assert(i[1].type == FAWKC_NUM);
					numargs = i[1].data.num;

					/* when in a for(in), there are extra SYMREF entries on the stack, remove them */
					while(STACKR(-1).type == FAWK_SYMREF) POP();

					cell = &STACKR(-1);
					if (cell->type != FAWK_NUM) { FAWK_ERROR(ctx, 64, (FAWK_ERR, "invalid symref; tip: array-in-array can not be created with [] in one step\n")); return FAWK_ER_ERROR; }
					ctx->ip = cell->data.num;
					POP();
					cell = &STACKR(-1);
					fawk_assert(cell->type == FAWK_NUM);
					ctx->fp = cell->data.num;
					POP();
					for(n = 0; n <= numargs; n++)
						POP();
					cell = PUSH(); if (cell == NULL) return FAWK_ER_ERROR;
					*cell = retval;

					cell = &STACKR(-2);
					if (cell->type == FAWK_CCALL_RET) {
						STACKR(-2) = STACKR(-1); /* move the result so it replaces the marker */
						STACKR(-1).type = FAWK_NIL;
						POP();
						return FAWK_ER_FIN;
					}
				}
				break;
			case FAWKI_JMP:
				exec_jump:;
				{
					int addr = i[1].data.num;
					fawk_assert(i[1].type == FAWKC_NUM);
					fawk_assert(addr >= 0);
					fawk_assert(addr < ctx->code.used);
					ctx->ip = addr-2; /* compensate for the ++ */
				}
				BREAK(1);
			case FAWKI_ABORT:
				FAWK_ERROR(ctx, 64, (FAWK_ERR, "libfawk abort"));
				if ((i[1].type == FAWKC_STR) || (i[1].type == FAWKC_CSTR))
					libfawk_error(ctx, i[1].data.cstr, "<runtime>", ctx->code.code[ctx->ip].line, 0);
				return FAWK_ER_ERROR;
		}
		ctx->ip++;
	}
	return FAWK_ER_STEPS;
}

FAWK_API int fawk_call1(fawk_ctx_t *ctx, const char *funcname)
{
	fawk_cell_t *cell, *func = htsp_get(&ctx->symtab, funcname);
	if ((func == NULL) || (func->type != FAWK_FUNC))
		return -1;
	cell = PUSH(); if (cell == NULL) return FAWK_ER_ERROR;
	cell->type = FAWK_CCALL_RET;
	cell = PUSH(); if (cell == NULL) return FAWK_ER_ERROR;
	cell->type = FAWK_SYMREF;
	cell->data.symref.is_local = cell->data.symref.idx_len = 0;
	cell->data.symref.idx = NULL;
	cell->data.symref.ref.global = func;
	return 0;
}

FAWK_API int fawk_call2(fawk_ctx_t *ctx, int argc)
{
	fawk_cell_t *funcref = &STACKR(-argc-1), *func = funcref->data.symref.ref.global;
	if ((funcref->type != FAWK_SYMREF) || (func->type != FAWK_FUNC) || ((func->data.func.numargs == func->data.func.numfixedargs) && (func->data.func.numargs < argc))) {
		while(argc > -2) { POP(); argc--; }
		return -1;
	}
	exec_call(ctx, argc);
	ctx->ip++;
	return 0;
}

